class MachineProvider(object):
    def get_all_shared_fate_zones(self):
        pass

    def list_available_machines(self, sf_zone):
        pass

    def usable(self):
        return False

    def decomission(self, machine):
        pass
